package mmc;

/**
 * medium definition
 */

public class Medium extends PhysicsModel{

    public int num;                // number of components
    public double[] Z;             // nucleus charge
    public double[] A;             // atomic number
    public double[] n;             // number of atoms in a molecule
    public double totZ;            // sum of charges of all nuclei

    public double ZA;              // <Z/A>
    public double I;               // ionization potential [eV]
    public double C1, C, a;        // ionization formula constants
    public double m, X0, X1, d0;   // ionization formula constants (continued)
    public double r;               // refraction index

    public double B[];             // radiation logarithm constant B
    public double P[];             // radiation logarithm constant bPrime

    public double rho;             // multiplicative density correction factor
    public double Ro;              // mass density [g/cm3]
    public double No;              // molecule density [number/cm3]
    public double M[];             // average nucleon weight in a nucleus [MeV]
    public String E[];             // element name
    public String name;            // medium name

    protected double ecut;         // cutoff energy [MeV]
    protected double vcut;         // relative cutoff energy
    private double vCut;           // relative cutoff energy - call setCut(E) to set this

    private Integral L;            // Needed for calculation of the
    public double[] mN;            // Woods-Saxon potential factor


    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize medium by its name and proposed values of energy cut (ecut) and fractional energy cut (vcut).
     * The cuts ecut and vcut to be used must satisfy ecut&gt;0 and 0&lt;=vcut&lt;=1; If both values satisfy these
     * inequalities, the lower from E*vcut and ecut will be used. If only one value satisfies the inequalities,
     * only that one value is used. If both values are outside these intervals, vcut=1 is assumed.
     */

    public Medium(String w, double ecut, double vcut, double rho){
	name=w;
	this.rho=rho>0?rho:1;
	if(w.equalsIgnoreCase("water")) initWater();
	else if(w.equalsIgnoreCase("ice")) initIce();
	else if(w.equalsIgnoreCase("standard rock")) initStandardrock();
	else if(w.equalsIgnoreCase("frejus rock")) initFrejusrock();
	else if(w.equalsIgnoreCase("iron")) initIron();
	else if(w.equalsIgnoreCase("hydrogen")) initHydrogen();
	else if(w.equalsIgnoreCase("lead")) initLead();
	else if(w.equalsIgnoreCase("uranium")) initUranium();
	else if(w.equalsIgnoreCase("air")) initAir();
	else {
	    Output.err.println("Warning (in Medium/Medium): unknown medium: \""+w+"\", defaulting to water");
	    name="water";
	    initWater();
	}
	this.ecut=ecut;
	this.vcut=vcut;
	vCut=1.;
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialization of arrays
     */

    private void inita(int i){
	num=i;
	Z = new double[num];
	A = new double[num];
	n = new double[num];
	B = new double[num];
	P = new double[num];
	M = new double[num];
	E = new String[num];
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialization code, common to all media
     */

    private void initr(){
	int i;
	boolean flag=false;
	double aux1=0, aux2=0;
	Ro*=rho;
	for(i=0;i<num;i++){
	    aux1+=n[i]*Z[i];
	    aux2+=n[i]*A[i];
	    B[i]=elB(i);
	    P[i]=elP(i);
	    M[i]=(Z[i]*Mp+(A[i]-Z[i])*Mn)/A[i];
	    if(Z[i]!=1) flag=true;
	}
	totZ=aux1;
	ZA=aux1/aux2;
	No=Ro*Na/aux2;
	C1=2*Log10;
	r=1.31;                         // only for ice - change if needed
	ecut=Me/Math.sqrt(1-1/(r*r));   // in order to emit Cerenkov photons
	if(flag){
	    mN = new double[num];
	    L = new Integral(iromb, imaxs, iprec);
	    for(i=0;i<num;i++) if(Z[i]!=1){
		r0=Math.pow(A[i], 1./3);
		r0=1.12*r0-0.86/r0;
		mN[i]=1-4*Pi*0.17*L.integrateWithSubstitution(r0 , -1, this, 2)/A[i];
	    }
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * set the value of radiation logarithm constant B
     */

    private double elB(int i){
	int z=(int)Math.round(Z[i]);
	switch(z){
	case 1: return 202.4;
	case 2: return 151.9;
	case 3: return 159.9;
	case 4: return 172.3;
	case 5: return 177.9;
	case 6: return 178.3;
	case 7: return 176.6;
	case 8: return 173.4;
	case 9: return 170.0;
	case 10: return 165.8;
	case 11: return 165.8;
	case 12: return 167.1;
	case 13: return 169.1;
	case 14: return 170.8;
	case 15: return 172.2;
	case 16: return 173.4;
	case 17: return 174.3;
	case 18: return 174.8;
	case 19: return 175.1;
	case 20: return 175.6;
	case 21: return 176.2;
	case 22: return 176.8;
	case 26: return 175.8;
	case 29: return 173.1;
	case 32: return 173.0;
	case 35: return 173.5;
	case 42: return 175.9;
	case 50: return 177.4;
	case 53: return 178.6;
	case 74: return 177.6;
	case 82: return 178.0;
	case 92: return 179.8;
	default: return 182.7;
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * set the value of radiation logarithm constant bPrime
     */

    private double elP(int i){
	int z=(int)Math.round(Z[i]);
	switch(z){
	case 1: return 446;
	default: return 1429;
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize water
     */

    private void initWater(){
	inita(2);
	E[0]="H";
	E[1]="O";
	Z[0]=1; // H
	Z[1]=8; // O
	A[0]=1.00794;
	A[1]=15.9994;
	n[0]=2;
	n[1]=1;
	I=75.0;
	C=-3.5017;
	a=0.09116;
	m=3.4773;
	X0=0.2400;
	X1=2.8004;
	d0=0;
	Ro=1.000;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize ice
     */

    private void initIce(){
	inita(2);
	E[0]="H";
	E[1]="O";
	Z[0]=1; // H
	Z[1]=8; // O
	A[0]=1.00794;
	A[1]=15.9994;
	n[0]=2;
	n[1]=1;
	I=75.0;
	C=-3.5017;
	a=0.09116;
	m=3.4773;
	X0=0.2400;
	X1=2.8004;
	d0=0;
	Ro=0.917;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize standard rock
     */

    private void initStandardrock(){
	inita(1);
	E[0]="Standard Rock";
	Z[0]=11; // Ionization potential and density corrections
	A[0]=22; // are close to those of calcium carbonate
	n[0]=1;
	I=136.4;
	C=-3.7738;
	a=0.08301;
	m=3.4120;
	X0=0.0492;
	X1=3.0549;
	d0=0;
	Ro=2.650;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize standard rock
     */

    private void initFrejusrock(){
	inita(1);
	E[0]="Frejus Rock";
	Z[0]=10.12;
	A[0]=20.34;
	n[0]=1;
	I=149.0;
	C=-5.053;
	a=0.078;
	m=3.645;
	X0=0.288;
	X1=3.196;
	d0=0;
	Ro=2.740;
	initr();
    }


    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize iron
     */

    private void initIron(){
	inita(1);
	E[0]="Fe";
	Z[0]=26;
	A[0]=55.845;
	n[0]=1;
	I=286.0;
	C=-4.2911;
	a=0.14680;
	m=2.9632;
	X0=-0.0012;
	X1=3.1531;
	d0=0.12;
	Ro=7.874;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize hydrogen
     */

    private void initHydrogen(){
	inita(1);
	E[0]="H";
	Z[0]=1;
	A[0]=1.00794;
	n[0]=1;
	I=21.8;
	C=-3.0977;
	a=0.13483;
	m=5.6249;
	X0=0.4400;
	X1=1.8856;
	d0=0;
	Ro=0.07080;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize lead
     */

    private void initLead(){
	inita(1);
	E[0]="Pb";
	Z[0]=82;
	A[0]=207.2;
	n[0]=1;
	I=823.0;
	C=-6.2018;
	a=0.09359;
	m=3.1608;
	X0=0.3776;
	X1=3.8073;
	d0=0.14;
	Ro=11.350;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize uranium
     */

    private void initUranium(){
	inita(1);
	E[0]="U";
	Z[0]=92;
	A[0]=238.0289;
	n[0]=1;
	I=890.0;
	C=-5.8694;
	a=0.19677;
	m=2.8171;
	X0=0.2260;
	X1=3.3721;
	d0=0.14;
	Ro=18.950;
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /*
     * initialize air
     */

    private void initAir(){
	final double fr1=2*78.1;
	final double fr2=2*21.0;
	final double fr3=0.9;
	final double fra=fr1+fr2+fr3;
	inita(3);
	E[0]="N";
	E[1]="O";
	E[2]="Ar";
	Z[0]=7; // N
	Z[1]=8; // O
	Z[2]=18; // Ar
	A[0]=14.0067;
	A[1]=15.9994;
	A[2]=39.948;
	n[0]=fr1/fra;
	n[1]=fr2/fra;
	n[1]=fr3/fra;
	I=85.7;
	C=-10.5961;
	a=0.10914;
	m=3.3994;
	X0=1.7418;
	X1=4.2759;
	d0=0;
	Ro=1.205e-3; // dry, 1 atm
	initr();
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * return the value of the fractional energy cut, as described in the help for the constructor
     */

    public double vCut(double E){
	double aux;
	aux=ecut/E;
	if(ecut>0){
	    if(vcut>0 && vcut<=1){
		if(aux<vcut) vCut=aux;
		else vCut=vcut;
	    }
	    else vCut=aux;
	}
	else{
	    if(vcut>0 && vcut<=1) vCut=vcut;
	    else vCut=1.;
	}
	return vCut;
    }

    //----------------------------------------------------------------------------------------------------//

    private double r0;

    /**
     * Woods-Saxon potential calculation - interface to Integral
     */

    public double function(double r){
	final double a=0.54;
	return r*r/(1+Math.exp((r-r0)/a));
    }

}
