package mmc;

/**
 * particle definition (muon)
 */

public class Particle extends PhysicsModel{

    public double r=0, x=0, y=0, z=0, t=0;             // coordinates [cm] and age [sec]
    public double theta=0, phi=0;                      // zenith and azimuth of the momentum in [deg]
    public double costh=1, sinth=0, cosph=1, sinph=0;  // cos and sin of theta and phi
    public double p=0, p2=0, e=0;                      // momentum, momentum square and energy in [MeV]
    public double m=Mmu, l=Lmu, c=1;                   // mass [MeV], lifetime [sec] and charge
    public String name="mu";                           // name of the particle
    public double low=m;                               // energy below which the particle is lost [MeV]
    public int type=1;                                 // particle type: 1 for muon, 2 for tau

    protected Scattering s;
    protected Propagate pr;

    private Integral I;

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize particle
     */

    public Particle(Propagate pr, String name){
	if(name.startsWith("tau")){
	    this.name="tau";
	    type=2;
	    m=Mtau;
	    l=Ltau;
	}
	else{
	    this.name="mu";
	    type=1;
	    m=Mmu;
	    l=Lmu;
	}
	low=m; e=m;
	this.pr=pr;
	if(elow>low) low=elow;
	if(ebig<100*low) ebig=bigEnergy;
	s = new Scattering(this);
 	I = new Integral(iromb, imaxs, iprec2);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * store particle information
     */

    public Particle(String name, double x, double y, double z, double theta, double phi, double e, double t){
	if(name.equals("tau") || name.equals("tau-") || name.equals("tau+")){
	    if(name.equals("tau+")) type=-33;
	    else type=-34;
	    m=Mtau;
	    l=Ltau;
	}
	else if(name.equals("mu") || name.equals("mu-") || name.equals("mu+")){
	    if(name.equals("mu+")) type=-5;
	    else type=-6;
	    m=Mmu;
	    l=Lmu;
	}
	else if(name.equals("e") || name.equals("e-") || name.equals("e+")){
	    if(name.equals("e+")) type=-2;
	    else type=-3;
	    m=Mmu;
	    l=Lmu;
	}
	else{
	    if(name.equals("delta")) type=-1002;
	    else if(name.equals("brems")) type=-1001;
	    else if(name.equals("munu")) type=-1004;
	    else if(name.equals("epair")) type=-1003;
	    else if(name.equals("hadr")) type=-1006;
	    m=0;
	    l=0;
	}
	low=m;
	setEnergy(e);
	location(name, t, x, y, z, theta, phi);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * initialize the location and direction of the particle, time in sec, x, y, z in cm, theta and phi in deg
     */

    public void location(String name, double time, double x, double y, double z, double theta, double phi){
	this.name=name;
	r=0;
	t=time;
	this.x=x;
	this.y=y;
	this.z=z;
	this.theta=theta;
	this.phi=phi;
	theta*=(Pi/180);
	phi*=(Pi/180);
	costh=Math.cos(theta);
	sinth=Math.sin(theta);
	cosph=Math.cos(phi);
	sinph=Math.sin(phi);
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * advances the particle by the given distance
     */

    public void advance(double dr, double ei, double ef){
	r+=dr;
	if(pr.exactTime) t+=getdt(ei, ef);
	else t+=dr/C;
	if(pr.molieScat){
	    double tho, max, rnd1, rnd2, sx, sy, sz, tx, ty, tz;
	    tho=s.gettho(dr, ei, ef);
	    max=1/sqrt2;

	    rnd1=pr.S.sndrn(Propagate.rand.nextDouble(), 0, tho, -max, max, false);
	    rnd2=pr.S.sndrn(Propagate.rand.nextDouble(), 0, tho, -max, max, false);

	    sx=(rnd1/sqrt3+rnd2)/2;
	    tx=rnd2;

	    rnd1=pr.S.sndrn(Propagate.rand.nextDouble(), 0, tho, -max, max, false);
	    rnd2=pr.S.sndrn(Propagate.rand.nextDouble(), 0, tho, -max, max, false);

	    sy=(rnd1/sqrt3+rnd2)/2;
	    ty=rnd2;

	    sz=Math.sqrt(Math.max(1-(sx*sx+sy*sy), 0));
	    tz=Math.sqrt(Math.max(1-(tx*tx+ty*ty), 0));

	    double ax, ay, az;

	    ax=sinth*cosph*sz+costh*cosph*sx-sinph*sy;
	    ay=sinth*sinph*sz+costh*sinph*sx+cosph*sy;
	    az=costh*sz-sinth*sx;

	    x+=ax*dr;
	    y+=ay*dr;
	    z+=az*dr;

	    ax=sinth*cosph*tz+costh*cosph*tx-sinph*ty;
	    ay=sinth*sinph*tz+costh*sinph*tx+cosph*ty;
	    az=costh*tz-sinth*tx;

	    costh=az;
	    sinth=Math.sqrt(Math.max(1-costh*costh, 0));
	    if(sinth!=0){
		sinph=ay/sinth;
		cosph=ax/sinth;
	    }

	    theta=Math.acos(costh>1?1:costh<-1?-1:costh)*180/Pi;
	    phi=Math.acos(cosph>1?1:cosph<-1?-1:cosph)*180/Pi;
	    if(sinph<0) phi=360-phi; if(phi>=360) phi-=360;

	}
	else{
	    x+=sinth*cosph*dr;
	    y+=sinth*sinph*dr;
	    z+=costh*dr;
	}
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * sets the energy of the particle
     */

    public void setEnergy(double e){
	this.e=e;
	p2=e*e-m*m;
	p=Math.sqrt(Math.max(p2, 0));
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * function for time delta calculation - interface to Integral
     */

    public double function(double E){
	final boolean DEBUG=false;
	double aux;
	aux=pr.s.function(E);
	aux*=e/(p*C);
	if(DEBUG) Output.err.println(" # "+Output.f(e)+" \t "+Output.f(aux));
	return aux;
    }

    //----------------------------------------------------------------------------------------------------//

    /**
     * time delta, corresponding to the given propagation distance
     */

    public double getdt(double ei, double ef){
	if(jt){
	    if(Math.abs(ei-ef)>Math.abs(ei)*halfPrecision){
		double aux=J.interpolate(ei);
		double aux2=aux-J.interpolate(ef);
		if(Math.abs(aux2)>Math.abs(aux)*halfPrecision) return aux2;
	    }
	    return Jdf.interpolate((ei+ef)/2)*(ef-ei);
	}
	else return I.integrateWithLog(ei, ef, this);
    }

    //----------------------------------------------------------------------------------------------------//

    boolean df=false;
    public Interpolate J;
    public Interpolate Jdf;
    public boolean jt=false;

    /**
     * 1d parametrization - interface to Interpolate
     */

    public double functionInt(double e){
	if(df) return function(e);
	else return I.integrateWithLog(e, low, this);
    }

}
